# core-image-tester
# Copyright (C) 2015 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""ADT Cloud worker functional code."""
import logging
import os
import shutil
import tarfile
import tempfile

from core_image_tester import (
    cloud,
    constants,
    utils,
)
from core_image_tester.queue import MessageActions


logger = logging.getLogger(__name__)


class CoreImageTesterWorker(object):

    def __init__(self, config, publisher):
        self.config = config
        self.publish_result = publisher
        self.worker_name = constants.HOSTNAME

    def __call__(self, body):
        """Process incomming test request.

        Run requested tests and posts results to the 'results' queue
        for later checking.
        """
        body['worker'] = self.worker_name
        extra = constants.LOGGING_EXTRA.copy()
        extra.update(body)
        # Run requested tests safely.
        tarball_path = None
        try:
            nova_image_id = body['nova_image_id']
            image_name = body['image_name']
            device = body['device']
            channel = body['channel']
        except KeyError as e:
            logger.error("Message does not contain required ", extra=extra)
            return MessageActions.Retry

        with tempfile.TemporaryDirectory() as test_dir:
            logger.info("Downloading test sources...", extra=extra)
            try:
                get_test_sources(test_dir, self.config['adt']['test_branch'])
            except utils.BetterCalledProcessError as e:
                logger.error("Unable to retrieve test sources: %s",
                             e, extra=extra)
                return MessageActions.Retry
            logger.info("Test sources downloaded OK.", extra=extra)

            with tempfile.TemporaryDirectory() as result_dir:
                logger.info("Running autopkgtest...", extra=extra)
                exit_code = run_autopkgtest(
                    result_dir,
                    test_dir,
                    nova_image_id,
                    self.config
                )
                try:
                    sanitize_results(result_dir, self.config)
                except IOError as e:
                    logger.error("Unable to sanitize results: %s",
                                 e, extra=extra)
                    return MessageActions.Retry
                body['exit_code'] = exit_code
                logger.info(
                    "Autopkgtest completed with exit code %s.",
                    exit_code,
                    extra=extra
                )

                tarball_path = tar_directory(
                    result_dir,
                    "core-{}".format(image_name)
                )
                try:
                    container_name = "core-{}-{}-{}".format(
                        channel,
                        device,
                        image_name,
                    ).replace('/', '_')
                    logger.info(
                        "Uploading results to swift container '%s'...",
                        container_name,
                        extra=extra
                    )
                    success = cloud.upload_to_swift(
                        tarball_path,
                        'results.tgz',
                        container_name,
                        self.config['nova']
                    )
                    if not success:
                        logger.error(
                            "Unable to upload results to swift container '%s'",
                            container_name,
                            extra=extra
                        )
                        return MessageActions.Retry
                    logger.info("Results uploaded to swift OK.", extra=extra)
                finally:
                    os.unlink(tarball_path)

        # code to run the tests isn't in place yet, so we're publishing
        # incomplete results, but this allows the result-checker work to
        # start:
        self.publish_result(body)
        logger.info('Done!', extra=extra)
        return MessageActions.Acknowledge


def get_test_sources(target_dir, test_branch):
    """Copy the test sources into the target_dir.

    'target_dir' should be an empty directory on disk. It must exist, and bzr
    will export the test sources into that directory.

    """
    utils.check_call(['bzr', 'export', target_dir, test_branch])


def run_autopkgtest(result_dir, test_dir, nova_image, config):
    """Run autopkgtest.

    Results will be stored in 'result_dir'.
    Tests will be taken from 'test_dir'.
    'nova_image' must contain the name of the nova image to use.
    'config' will be read for the nova flavour and other nova arguments.

    This function will return the adt-run exit code.

    """
    adt_env = {
        k.upper(): v for k, v in config['nova'].items() if k.startswith('os')
    }
    adt_args = [
        'adt-run',
        # Options for adt-run itself:
        '-d',
        '--built-tree', test_dir,
        '--output-dir', result_dir,
        # Specify the ssh backend, and the uci-nova setup script:
        '---', 'ssh', '-s', 'uci-nova', '-d', '--',
        '-d',
        '--flavor', config['adt']['image_flavor'],
        '--image', nova_image,
        '--console', os.path.join(result_dir, 'nova-console.log'),
    ]
    if 'extra_args' in config['nova']:
        adt_args.append(config['nova']['extra_args'])

    try:
        utils.check_call(adt_args, env=adt_env)
    except utils.BetterCalledProcessError as e:
        return e.returncode
    return 0


def sanitize_results(results_dir, config):
    """Filter out any secrets exposed in select log files.

    Filters log files stored in 'results_dir'.
    'config' will be read for the nova secrets used by run_autopkgtest.
    """
    secrets = [
        v for k, v in config.items() if k.startswith('os')
    ]
    sanitation_logs = ['log', 'nova-console.log', 'summary']
    for log in sanitation_logs:
        log_path = os.path.join(results_dir, log)
        if not os.path.isfile(log_path):
            continue
        _unused, temp_path = tempfile.mkstemp()
        with open(log_path) as in_file, open(temp_path, 'w') as out_file:
            for line in in_file:
                for secret in secrets:
                    line = line.replace(secret, '<REDACTED>')
                out_file.write(line)
        shutil.move(temp_path, log_path)


def tar_directory(directory_to_tar, tarball_root):
    """Create a gzip'd tarball of the given directory.

    :returns: the path to the newly created tarball. The caller is responsible
    for deleting the tarball.

    """
    # Compress the results directory in a temporary file.
    _unused, tarball_path = tempfile.mkstemp()
    with tarfile.open(tarball_path, "w:gz") as tar:
        for entry in os.listdir(directory_to_tar):
            tar.add(
                os.path.join(directory_to_tar, entry),
                arcname=os.path.join(tarball_root, entry),
            )
    return tarball_path
